@@ -16,7 +16,7 @@
  */
 
 /*
- * $Id: DOMElementImpl.cpp 678766 2008-07-22 14:00:16Z borisk $
+ * $Id$
  */
 
 #include "DOMElementImpl.hpp"
@@ -43,7 +43,7 @@ XERCES_CPP_NAMESPACE_BEGIN
 class DOMAttr;
 
 DOMElementImpl::DOMElementImpl(DOMDocument *ownerDoc, const XMLCh *eName)
-    : fNode(ownerDoc), fParent(ownerDoc), fAttributes(0), fDefaultAttributes(0)
+    : fNode(this, ownerDoc), fParent(this, ownerDoc), fAttributes(0), fDefaultAttributes(0)
 {
     DOMDocumentImpl *docImpl = (DOMDocumentImpl *)ownerDoc;
     fName = docImpl->getPooledString(eName);
@@ -60,8 +60,8 @@ DOMElementImpl::DOMElementImpl(DOMDocument *ownerDoc, const XMLCh *eName)
 
 DOMElementImpl::DOMElementImpl(const DOMElementImpl &other, bool deep)
     : DOMElement(other),
-      fNode (other.fParent.fOwnerDocument),
-      fParent (other.fParent.fOwnerDocument),
+      fNode(this, other.fParent.fOwnerDocument),
+      fParent(this, other.fParent.fOwnerDocument),
       fAttributes(0),
       fDefaultAttributes(0)
 {
@@ -477,7 +477,17 @@ void DOMElementImpl::release()
     DOMDocumentImpl* doc = (DOMDocumentImpl*) fParent.fOwnerDocument;
     if (doc) {
         fNode.callUserDataHandlers(DOMUserDataHandler::NODE_DELETED, 0, 0);
+        // release children
         fParent.release();
+        // release attributes
+        fAttributes->hasDefaults(false);
+        XMLSize_t count;
+        while((count = fAttributes->getLength()) != 0)
+        {
+            DOMNode* attr = fAttributes->removeNamedItemAt(count-1);
+            attr->release();
+        }
+
         doc->release(this, DOMMemoryManager::ELEMENT_OBJECT);
     }
     else {
@@ -569,7 +579,10 @@ const XMLCh* DOMElementImpl::getBaseURI() const
            const XMLCh*     DOMElementImpl::lookupNamespaceURI(const XMLCh* prefix) const  {return fNode.lookupNamespaceURI(prefix); }
            void*            DOMElementImpl::getFeature(const XMLCh* feature, const XMLCh* version) const {return fNode.getFeature(feature, version); }
 
-
+// Macro-in implementation accessors.
+DOMNODEIMPL_IMPL(DOMElementImpl);
+DOMPARENTIMPL_IMPL(DOMElementImpl);
+DOMCHILDIMPL_IMPL(DOMElementImpl);
 
 bool DOMElementImpl::isEqualNode(const DOMNode* arg) const
 {
@@ -672,4 +685,199 @@ const DOMTypeInfo *DOMElementImpl::getSchemaTypeInfo() const
     return &DOMTypeInfoImpl::g_DtdValidatedElement;
 }
 
+// DOMElementTraversal
+DOMElement * DOMElementImpl::getFirstElementChild() const
+{
+    DOMNode* n = getFirstChild();
+    while (n != NULL) {
+        switch (n->getNodeType()) {
+            case DOMNode::ELEMENT_NODE:
+                return (DOMElement*) n;
+            case DOMNode::ENTITY_REFERENCE_NODE:
+                {
+                    DOMElement* e = getFirstElementChild(n);
+                    if (e != NULL)
+                        return e;
+                }
+                break;
+            default:
+                break;
+        }
+        n = n->getNextSibling();
+    }
+    return NULL;
+}
+
+DOMElement * DOMElementImpl::getLastElementChild() const
+{
+    DOMNode* n = getLastChild();
+    while (n != NULL) {
+        switch (n->getNodeType()) {
+            case DOMNode::ELEMENT_NODE:
+                return (DOMElement*) n;
+            case DOMNode::ENTITY_REFERENCE_NODE:
+                {
+                    DOMElement* e = getLastElementChild(n);
+                    if (e != NULL)
+                        return e;
+                }
+                break;
+            default:
+                break;
+        }
+        n = n->getPreviousSibling();
+    }
+    return NULL;
+}
+
+DOMElement * DOMElementImpl::getNextElementSibling() const
+{
+    DOMNode* n = getNextLogicalSibling(this);
+    while (n != NULL) {
+        switch (n->getNodeType()) {
+            case DOMNode::ELEMENT_NODE:
+                return (DOMElement*) n;
+            case DOMNode::ENTITY_REFERENCE_NODE:
+                {
+                    DOMElement* e = getFirstElementChild(n);
+                    if (e != NULL)
+                        return e;
+                }
+                break;
+            default:
+                break;
+        }
+        n = getNextLogicalSibling(n);
+    }
+    return NULL;
+}
+
+DOMElement * DOMElementImpl::getPreviousElementSibling() const
+{
+    DOMNode* n = getPreviousLogicalSibling(this);
+    while (n != NULL) {
+        switch (n->getNodeType()) {
+            case DOMNode::ELEMENT_NODE:
+                return (DOMElement*) n;
+            case DOMNode::ENTITY_REFERENCE_NODE:
+                {
+                    DOMElement* e = getLastElementChild(n);
+                    if (e != NULL)
+                        return e;
+                }
+                break;
+            default:
+                break;
+        }
+        n = getPreviousLogicalSibling(n);
+    }
+    return NULL;
+}
+
+XMLSize_t DOMElementImpl::getChildElementCount() const
+{
+    XMLSize_t count = 0;
+    DOMElement* child = getFirstElementChild();
+    while (child != NULL) {
+        ++count;
+        child = child->getNextElementSibling();
+    }
+    return count;
+}
+
+// Returns the first element node found from a
+// non-recursive in order traversal of the given node.
+DOMElement* DOMElementImpl::getFirstElementChild(const DOMNode* n) const
+{
+    const DOMNode* top = n;
+    while (n != NULL) {
+        if (n->getNodeType() == DOMNode::ELEMENT_NODE) {
+            return (DOMElement*) n;
+        }
+        DOMNode* next = n->getFirstChild();
+        while (next == NULL) {
+            if (top == n) {
+                break;
+            }
+            next = n->getNextSibling();
+            if (next == NULL) {
+                n = n->getParentNode();
+                if (n == NULL || top == n) {
+                    return NULL;
+                }
+            }
+        }
+        n = next;
+    }
+    return NULL;
+}
+
+// Returns the first element node found from a
+// non-recursive reverse order traversal of the given node.
+DOMElement* DOMElementImpl::getLastElementChild(const DOMNode* n) const
+{
+    const DOMNode* top = n;
+    while (n != NULL) {
+        if (n->getNodeType() == DOMNode::ELEMENT_NODE) {
+            return (DOMElement*) n;
+        }
+        DOMNode* next = n->getLastChild();
+        while (next == NULL) {
+            if (top == n) {
+                break;
+            }
+            next = n->getPreviousSibling();
+            if (next == NULL) {
+                n = n->getParentNode();
+                if (n == NULL || top == n) {
+                    return NULL;
+                }
+            }
+        }
+        n = next;
+    }
+    return NULL;
+}
+
+// Returns the next logical sibling with respect to the given node.
+DOMNode* DOMElementImpl::getNextLogicalSibling(const DOMNode* n) const
+{
+    DOMNode* next = n->getNextSibling();
+    // If "n" has no following sibling and its parent is an entity reference node we
+    // need to continue the search through the following siblings of the entity
+    // reference as these are logically siblings of the given node.
+    if (next == NULL) {
+        DOMNode* parent = n->getParentNode();
+        while (parent != NULL && parent->getNodeType() == DOMNode::ENTITY_REFERENCE_NODE) {
+            next = parent->getNextSibling();
+            if (next != NULL) {
+                break;
+            }
+            parent = parent->getParentNode();
+        }
+    }
+    return next;
+}
+
+// Returns the previous logical sibling with respect to the given node.
+DOMNode* DOMElementImpl::getPreviousLogicalSibling(const DOMNode* n) const
+{
+    DOMNode* prev = n->getPreviousSibling();
+    // If "n" has no previous sibling and its parent is an entity reference node we
+    // need to continue the search through the previous siblings of the entity
+    // reference as these are logically siblings of the given node.
+    if (prev == NULL) {
+        DOMNode* parent = n->getParentNode();
+        while (parent != NULL && parent->getNodeType() == DOMNode::ENTITY_REFERENCE_NODE) {
+            prev = parent->getPreviousSibling();
+            if (prev != NULL) {
+                break;
+            }
+            parent = parent->getParentNode();
+        }
+    }
+    return prev;
+}
+
+
 XERCES_CPP_NAMESPACE_END
